
function filter(arr, func)
    local new_index = 1
    local size_orig = #arr
    for old_index, v in ipairs(arr) do
        if func(v, old_index) then
            arr[new_index] = v
            new_index = new_index + 1
        end
    end
    for i = new_index, size_orig do arr[i] = nil end
end


CardAttack = class("CardAttack", Weapon)

function CardAttack:init(player, cards)
    Weapon.init(self, player)
    self:setCards(cards or {})
    self.executions = {}
end

function CardAttack:setCards(cards)
    self.cards = cards
    
    local cost = 0
    for i = #cards, 1, -1 do
        cost = cards[i]:calculateCost(cost)
    end
    
    self.stats.cooldown.base = cost
end

function CardAttack:fire()
    Weapon.fire(self)
    table.insert(self.executions, CardExecution(self.player, self.cards))
end

function CardAttack:draw() end

function CardAttack:update(dt)
    if #self.cards > 0 then
        Weapon.update(self, dt)
    end
    
    filter(self.executions, function(exec)
        return not exec:update(dt)
    end)
end

--- CardList: class
CardList = class("CardList")

function CardList:init(cards)
    self.cards = {}
    for _, card in pairs(cards) do
        self:insert(card)
    end
end

function CardList:count()
    return #self.cards
end

function CardList:insert(card)
    table.insert(self.cards, card)
end

function CardList:clone()
    return CardList(self.cards)
end

--- CardExecution: class
CardExecution = class("CardExecution")

EXEC_NEXT = 1
EXEC_RUNNING = 2
EXEC_SPAWN = 3
EXEC_SPLIT = 4

function clone(t)
    if t.clone then
        return t:clone()
    end
    local cloned = {}
    for k, v in pairs(t) do
        cloned[k] = v
    end
    return cloned
end

function CardExecution:init(owner, cards)
    self.owner = owner
    self.cards = cards
    self.index = 1
    self.data = {}
end

function CardExecution:isComplete()
    return self.index > #self.cards
end

function CardExecution:advance()
    self.index = self.index + 1
    if not self:isComplete() then
        self.cards[self.index]:start(self.owner, self)
    end
end

function CardExecution:update(dt)
    if self:isComplete() then return true end
        
    local result, newOwner = self.cards[self.index]:update(self.owner, self, dt)
    if result == EXEC_NEXT then
        self:advance()
    elseif result == EXEC_SPAWN then
        if newOwner then            
            self.owner = newOwner
            newOwner.execution = self
            self:advance()
        end
        return true
    elseif result == EXEC_SPLIT then
    end
    return self:isComplete()
end

function CardExecution:clone()
    c = CardExecution(self.owner, self.cards)
    c.index = self.index
    c.data = clone(self.data)
    return c
end

--- Card: class
Card = class("Card")

function Card:init()
    self.cost = 0.5
end

function Card:calculateCost(current)
    return current + self.cost
end

function Card:start(owner, exec)
end

function Card:update(owner, exec, dt)
    return EXEC_COMPLETE
end

--- SpawnCard: class
SpawnCard = class("SpawnCard", Card)

function SpawnCard:init()    
    self.settings = {follow = true, angleOffset = -90, duration = 5}
end

function SpawnCard:update(owner, exec, dt)
    -- spawn the thing    
    return EXEC_SPAWN, self:spawn(owner, exec)
end

function SpawnCard:spawn(player, exec)
    local ent = player.gameplay.scene:entity()
    local cast = ent:add(Cast, player, self.settings)
    return cast
end

--- SwordCard: class
SwordCard = class("SwordCard", SpawnCard)

function SwordCard:init()
    self.settings = 
    {
        sprite = tileset.sword:sprite(),
        follow = true, 
        angleOffset = -90, 
        duration = 1.5,
        pierce = 3
    }
    self.cost = 2    
    self.title = "Sword"
end

--- HomingCard: class
HomingCard = class("HomingCard", Card)

function HomingCard:init(radius)
    self.radius = radius or 5
    self.force = 10
    self.cost = 1
    self.title = "Homing"
end

function HomingCard:update(owner, exec, dt)
    if is(owner, Cast) then
        local x, y, w, h = owner.entity.x - self.radius, owner.entity.y - self.radius, self.radius * 2, self.radius * 2        
        local results = owner.gameplay.scene.world2d:queryBoxAll(x, y, w, h, ENEMY_CAT)
        if #results > 0 then
            local r = results[math.random(#results)]
            local enemy = r.body.entity:get(Monster)
            if enemy then
                exec.data.target = enemy
                owner.settings.follow = false
                owner.settings.tracking = true
                owner.settings.trackingForce = self.force
                owner.settings.target = enemy
                return EXEC_NEXT
            end
        end
    end
    return EXEC_RUNNING
end

--- LaunchCard: class
LaunchCard = class("LaunchCard", Card)

function LaunchCard:init()
    Card.init(self)
    self.title = "Launch"
end

function LaunchCard:update(owner, exec, dt)
    if is(owner, Cast) then
        owner.settings.follow = false
        owner.settings.gravity = 0
        local d = vec2(owner.entity.up.y, -owner.entity.up.x)
        owner.body:applyForce(d * 300)
    end
    return EXEC_NEXT
end

--- ThrowCard: class
ThrowCard = class("ThrowCard", Card)

function ThrowCard:init()
    Card.init(self)
    self.title = "Throw"
end

function ThrowCard:update(owner, exec, dt)
    if is(owner, Cast) then
        owner.settings.follow = false
        owner.settings.gravity = 1
        --local d = vec2(owner.dir.x, 2).normalized
        local d = vec2(owner.entity.up.y, -owner.entity.up.x)
        owner.body:applyForce(d * 500)
        owner.body:applyTorque(100 * -math.sign(d.x), true)
    end
    return EXEC_NEXT
end

--- TimerCard: class
TimerCard = class("TimerCard", Card)

function TimerCard:init()
    Card.init(self)
    self.title = "Timer"
end

function TimerCard:start(owner, exec)    
    exec.data.time = 1
    owner.time = owner.time + 1
end

function TimerCard:update(owner, exec, dt)
    if exec.data.time > 0 then
        exec.data.time = exec.data.time - dt
        if exec.data.time <= 0 then
            return EXEC_NEXT
        end
    end
    return EXEC_RUNNING
end

--- SizeCard: class
SizeCard = class("SizeCard", Card)

function SizeCard:init(growth)
    self.growth = growth or 1.25
    self.cost = 1
    self.title = "Size " .. (growth > 1 and "Up" or "Down" )
end

function SizeCard:update(owner, exec, dt)
    owner.settings.scale = owner.settings.scale * self.growth
    return EXEC_NEXT
end

SplitCard = class("SplitCard", Card)

function SplitCard:init(factor, spread)
    self.factor = factor or 3
    self.spread = spread or 45
    self.cost = 0.25
    self.title = "Split"
end

function SplitCard:cost(current)
    return current * (1 + self.cost * self.factor)
end

function SplitCard:update(owner, exec, dt)
    if is(owner, Cast) then
        local copies = {owner}
        exec:advance()
        
        local v = owner.body.linearVelocity
        local vlen = v.length
        local a = math.deg(math.atan(v.y / vlen, v.x / vlen))
        
        for i = 1, self.factor-1 do
            local ent = owner.gameplay.scene:entity()
            local settings = clone(owner.settings)
            local newExec = exec:clone()
            local cast = ent:add(Cast, owner.owner, settings, newExec)
            cast.time = owner.time
            cast.body.position = owner.body.position
            cast.entity.position = owner.entity.position
            cast.body.linearVelocity = v
            table.insert(copies, cast)
        end
        
        if owner.settings.follow then
            local delta = self.spread / (#copies)
            local angle = -self.spread / 2
            for _, cast in pairs(copies) do                
                cast.settings.aimOffset = angle
                angle = angle + delta                
            end
        else 
            local delta = self.spread / (#copies-1)
            local angle = -self.spread / 2
            for _, cast in pairs(copies) do                
                cast:setVelocityAngle(a + angle)
                angle = angle + delta
            end
        end 
        
        return EXEC_SPLIT
    end
    return EXEC_NEXT
end

--- Cast: class
Cast = class("Cast")

function Cast:created(owner, settings, exec)    
    self.owner = owner
    self.gameplay = self.owner.gameplay
    self.execution = exec
    if self.execution then
        self.execution.owner = self
    end
        
    self.settings =
    {
        size = settings.size or 0.5,
        density = settings.densize or 1,
        duration = settings.duration or 2,
        gravity = settings.gravity or 0,
        category = settings.category or WEAPON_CAT,
        mask = settings.mask or ENEMY_CAT,
        sensor = settings.sensor or true,
        sprite = settings.sprite or tileset.sword:sprite(),
        follow = settings.follow or false,
        angleOffset = settings.angleOffset or 0,
        aimOffset = settings.aimOffset or 0,
        pierce = settings.pierce or 1,
        damage = settings.damage or 10,
        knockback = settings.knockback or 5,
        tracking = false,
        trackingForce = 0,
        scale = settings.scale or 1
    }
        
    self.spr = self.entity:child()
    self.spr.sprite = self.settings.sprite
    self.spr.rz = self.settings.angleOffset
    
    self.entity.x = owner.entity.x
    self.entity.y = owner.entity.y
    local body = self.entity:add(physics2d.body, DYNAMIC)
    body.gravityScale = self.settings.gravity
    self.body = body
    local collider = self.entity:add(physics2d.circle, self.settings.size)
    collider.category = self.settings.category
    collider.mask = self.settings.mask
    collider.sensor = self.settings.sensor
    self.collider = collider
    --self.entity.sprite = self.settings.sprite

    self.aim = vec2(self.owner.aim.x, self.owner.aim.y)
    self.time = self.settings.duration
end

function Cast:collisionBegan2d(collision)
    local mob = collision.otherBody.entity and collision.otherBody.entity:get(Mob)
    local settings = self.settings
    if mob and settings.pierce ~= 0 then
        mob:applyDamage(self, settings.damage)
        mob:applyKnockback(self.entity.x, self.entity.y, settings.knockback)
        settings.pierce = settings.pierce - 1
        if settings.pierce == 0 then
            self.entity:destroy(0.01)
            return
        end
    end
end

function Cast:setVelocityAngle(angle)
    local vlen = self.body.linearVelocity.length    
    local a = math.rad(angle)
    local v = vec2(math.cos(a) * vlen, math.sin(a) * vlen)
    self.body.linearVelocity = v
    self.body.rotation = angle + self.settings.angleOffset
    self.entity.rz = self.body.rotation
end

function Cast:update(dt)
    local settings = self.settings
    
    if self.time > 0 then
        self.time = self.time - dt
        if self.time <= 0 then
            self.entity:destroy()
        end
    end
    
    self.body.gravityScale = settings.gravity
    
    if self.scale ~= settings.scale then
        self.entity.uniformScale = settings.scale
        self.collider.radius = settings.size * settings.scale
        self.scale = settings.scale
    end
    
    if settings.follow then
        self.aim.x = self.owner.aim.x
        self.aim.y = self.owner.aim.y
        if self.owner.mob then
            local d = self.aim.normalized
            local ao = math.rad(settings.aimOffset)
            local cs = math.cos(ao)
            local sn = math.sin(ao)            
            local dx = d.x * cs - d.y * sn
            local dy = d.x * sn + d.y * cs
            
            local angle = math.deg(math.atan(dy, dx))
            self.entity.x = self.owner.entity.x + dx
            self.entity.y = self.owner.entity.y + dy
            self.entity.rz = angle
            self.body.position = vec2(self.entity.x, self.entity.y)
            self.body.rotation = angle            
        end
    end
    
    if settings.tracking and settings.trackingForce > 0 and settings.target then
        if settings.target.entity.valid then
            dir = self.body.position - settings.target.body.position
            if dir.length > 0.01 then
                dir:normalize()            
                self.aim = self.body.linearVelocity.normalized
                if self.aim.length > 0 then
                    local angle = math.deg(math.atan(self.aim.y, self.aim.x))
                    self.body.rotation = angle
                end
                self.body:applyForce(dir * -settings.trackingForce)
            end
        else
            settings.tracking = false
            settings.target = nil
        end            
    end
    
    if self.execution then
        if self.execution:update(dt) then
            self.execution = nil
        end
    end
end


-- Stacks?
-- Fusions?

-- Spawn a thing?
-- Clone a thing?

-- Execution

-- Deck
-- CardList Table<Card>
-- SpawnCard <- Card
-- DaggerCard <- SpawnCard
-- SplitExecution
-- Shoot
-- Delay
-- Split
-- Nova
-- Area

