-- List container
List = require 'pl.List'

-- Game state management class (menus, editor, gameplay, etc...)
GameState = class('GameState')

function GameState:init()
    self.scene = scene.default2d()
    self.scene.time.autoUpdate = false
    self.scene.pixelsPerUnit = 16
    self.cam = self.scene.camera:get(camera)

end

function GameState:added(game)
    self:layoutUI(WIDTH or 1024, HEIGHT or 768)
end

function GameState:removed(game) end

function GameState:draw()
    self.scene:draw()
end

function GameState:update(dt)
    self.scene:update(dt)
end

function GameState:touched(touch)
    self.scene:touched(touch)
end

function GameState:sizeChanged(w, h)
    --self:layoutUI(w, h)
end

function GameState:getOrthoSize(w, h)
    local aspect = (w+0.0)/h
    local orthoSizeY = self.scene.camera:get(camera).orthoSize
    local orthoSizeX = orthoSizeY * aspect
    return orthoSizeX, orthoSizeY
end

function GameState:layoutUI(w, h)
    local safeArea = viewer.safeArea
    local orientation = viewer.orientation
    local horizontal = viewer.horizontal
    local vertical = viewer.vertical
    local idiom = viewer.idiom
    local pad = idiom == viewer.PAD
    local phone = idiom == viewer.PHONE
    local mac = idiom == (viewer.MAC or -1)
    local landscape = (orientation == viewer.LANDSCAPE_LEFT) or (orientation == viewer.LANDSCAPE_RIGHT)
    
    local canvasScale = 1.0
    local aspect = (w+0.0)/h
    
    if phone then
        canvasScale = 1.65
    end
    
    local orthoSizeY = 20
    local orthoSizeX = 20
    
    if phone then
        orthoSizeX = 13
        orthoSizeY = 13
    end
    
    if landscape then
        orthoSizeX = orthoSizeY * aspect
    else
        orthoSizeY = orthoSizeX / aspect
    end
    
    self.scene.camera.position = vec3(0, 0, -1)
    self.scene.camera:get(camera).orthoSize = orthoSizeY
    
    self.scene.canvas.scale = canvasScale
end

function GameState:addBackButton(y)
    local backButton = createButton(self.scene, "Back", 100, 65, ui.resources.rightRounded)
    self.backButton = backButton
    backButton:anchor(RIGHT, TOP)
    backButton.pivot = vec2(1, 1)
    backButton.y = (y or 0) - 50
    
    function backButton.tapped(entity, btn)
        self.game:pop()
    end
end

Game = class('Game')

function Game:init(initial)
    self.states = List{}
    if initial then self:push(initial) end
end

function Game:top(offset)
    offset = offset or 0
    return self.states[self.states:len() + offset]
end

function Game:push(state)
    if self.states:len() > 0 then
        self:top():removed(self)
    end    
    self.states:append(state)
    state.game = self
    state:added(self)
    --state:layoutUI(WIDTH, HEIGHT)
end

function Game:replace(state)
    if self.states:len() > 0 then self:pop() end
    self:push(state)
end

function Game:set(state)
    self.states:clear()
    self:push(state)
end

function Game:draw()
    local top = self:top()
    if top.drawPreviousState and self.states:len() >= 2 then
        self:top(-1):draw()
    end
    top:draw()
end

function Game:update(dt)
    local top = self:top()
    if top.updatePreviousState and self.states:len() >= 2 then
        self:top(-1):update(dt)
    end    
    self:top():update(dt)
end

function Game:touched(touch)
    local top = self:top()
    if not top:touched(touch) and top.touchPreviousState and self.states:len() >= 2 then
        self:top(-1):touched(touch)
    end
end

function Game:sizeChanged(w, h)
    self:top():sizeChanged(w, h)
end

function Game:pop(n)
    n = n or 1
    for i = 1, n do
        if self.states:len() == 0 then return end
        local state = self.states:pop()
        state:removed(self)
    end
    
    if self.states:len() > 0 then
        self:top():added(self)
        self:top():layoutUI(WIDTH, HEIGHT)
    end
end