require 'GameState'

--- Gameplay: class
-- Gameplay state class, handling all gameplay
Gameplay = class("Gameplay", GameState)

PLAYER_CAT = 1
ENEMY_CAT = 2
ITEM_CAT = 4
WEAPON_CAT = 8

function Gameplay:init(map)
    GameState.init(self)
    
    self.map = map

    self.map.world2d = self.scene.world2d
    self.mapEntity = self.scene:entity()    
    
    --self.scene.physics2d.debugDraw = true
    
    -- Adjust camera ortho size (zoom in/out)
    self.cam.orthoSize = 13
    self.camera = self.scene.camera
    
    function self.mapEntity.draw(entity)
        self.map:draw()
    end
    
    self.layers = 
    {
        self.map.ground,
        self.map.details,
        self.map.objects
    }
        
    -- Used to convert certain tiles into objects for gameplay
    local tileToObject = 
    {
        [tileset.player.id] = { class = Player, args = {} },
    }
    
    self.layers[3]:visit(function(x, y, z, id)
        local objectType = tileToObject[id]
        if objectType then
            self:spawnFromTile(objectType, x, y)
            self.layers[3]:set(x, y, 0)
        end
    end) 
    
    -- Game state variables
    self.timer = 0
    self.waveTimer = 0
    self.spawnTimer = 0
    self.waveTime = 60
    self.maxTime = 60 * 10
    self.wave = 1
    self.maxEnemies = 300
    self.enemyCount = 0
    
    self.waves = WAVES
    self:beginWave(self.waves[1])
end

function Gameplay:layoutUI()
end

function Gameplay:draw()
    GameState.draw(self)
    matrix.ortho()
    
    -- Draw player HUD
    if self.player then
        self.player:drawUI()
        
        -- Draw XP bar
        style.push().stroke(255, 185, 0).strokeWidth(2).fill(0)
        rect(10, HEIGHT-50, WIDTH - 20, 30)
        local p = self.player.xp / self.player.nextLevelXP
        style.noStroke().fill(228, 27, 165)
        rect(11, HEIGHT-49, (WIDTH - 22) * p, 28)
        
        style.fill(255).textAlign(CENTER|MIDDLE).fontSize(20)
        text(string.format("LV %d", self.player.level), WIDTH-75, HEIGHT - 32)
        
        -- Draw timer / wave info
        style.fill(255).textAlign(CENTER).fontSize(30)
        local minutes = math.floor(self.timer / 60)
        local seconds = math.floor(self.timer - minutes * 60)
        text(string.format("%2d:%2d / %d / %d", minutes, seconds, self.wave, self.enemyCount), WIDTH/2, HEIGHT-70)
        style.pop()
        
        -- Draw items on HUD
        
        --[[
        -- Weapons        
        style.spriteMode(CORNER)
        matrix.push().translate(10, HEIGHT - 100)        
        local itemSlot = tileset.itemSlot:sprite()
        local weapons = self.player.weapons
        for i = 1,6 do
            sprite(itemSlot, 0, 0, 32)
            if weapons[i] and weapons[i].icon then
               sprite(weapons[i].icon, 0, 0, 32) 
            end
            matrix.translate(32 + 4, 0)
        end

        -- Passive items
        matrix.pop().push().translate(10, HEIGHT - 100 - 32 - 4)        
        local passives = self.player.passives
        for i = 1,6 do
            sprite(itemSlot, 0, 0, 32)
            if passives[i] and passives[i].icon then
                sprite(passives[i].icon, 0, 0, 32) 
            end
            matrix.translate(32 + 4, 0)
        end
        
        matrix.pop()
        ]]
    end
    
end

--- Gameplay:despawnEnemies Despawn enemies within a box region
--- x: number
--- y: number
--- w: number
--- h: number
function Gameplay:despawnEnemies(x, y, w, h)
    local results = self.scene.world2d:queryBoxAll(x, y, w, h, ENEMY_CAT)
    for k,v in pairs(results) do
        if v.body.entity then
            local mob = v.body.entity:get(Mob)
            if mob and mob ~= self.player.mob then
                mob.entity:destroy()
            end
        end
    end
end

--- Gameplay:finish Called when the level waves have been completed, spawns the final boss
function Gameplay:finish()
    local results = self.scene.world2d:queryBoxAll(-100000, -100000, 200000, 200000, ENEMY_CAT | ITEM_CAT)
    self:spawn(DEATH, self:randomSpawnPoint())
    for k,v in pairs(results) do
        if v.body.entity then
            local mob = v.body.entity:get(Mob)
            if mob and mob ~= self.player.mob then
                mob:applyDamage(self.player, 1000000, true)
            end
            local gem = v.body.entity:get(Gem)
            if gem then
                gem.entity:destroy()
            end            
        end
    end
end

function Gameplay:waveLogic(dt)
    -- Handle wave logic
    self.timer = self.timer + dt
    self.waveTimer = self.waveTimer + dt
    if self.waveTimer > self.waveTime then
        -- WAVE END        
        if self.waves[self.wave+1] then
            self:beginWave(self.waves[self.wave+1])
            self.wave = self.wave + 1
            self.waveTimer = 0
        elseif not self.gameover then
            self.gameover = true
            self:finish()
        end
    else
        wave = self.waves[self.wave]
        if wave then
            self:updateWave(wave, dt)
        end
    end    
end

--- Gameplay:update Update the gameplay state, handling level logic such as waves
--- dt: number, the delta time
function Gameplay:update(dt)
    GameState.update(self, dt)
        
    local sx, sy = self:getOrthoSize(WIDTH, HEIGHT)
    if self.player then
        local cx = self.player.entity.x
        local cy = self.player.entity.y
        cx = math.min(math.max(cx, sx/2), MAP_SIZE+sx/2)
        cy = math.min(math.max(cy, sy/2), MAP_SIZE+sy/2)
        self.camera.x = cx + self.player.cameraOffset.x
        self.camera.y = cy + self.player.cameraOffset.y
    end
    
    -- Despawn enemies that are too far off screen (excluding bosses?)
    sx = sx * 1.5
    sy = sy * 1.5
    local x, y = self.player.entity.x, self.player.entity.y
    self:despawnEnemies(x - sx, y - sy, 1, sy*2) -- left
    self:despawnEnemies(x + sx, y - sy, 1, sy*2) -- right
    self:despawnEnemies(x - sx, y - sy, sx * 2, 1) -- bottom
    self:despawnEnemies(x - sx, y + sy, sx * 2, 1) -- bottom
end

--- Gameplay:showDamageNumber Spawn a damage number on screen
--- x: number
--- y: number
--- amount: number
function Gameplay:showDamageNumber(x, y, amount)
    amount = math.ceil(amount) + math.random(0, 5)
    local msg = self.scene:entity()
    msg:destroy(1)
    msg.x = x
    msg.y = y + 0.5
    msg.timer = 1
    
    function msg:draw()            
        style.push().fill(255, 255, 0, 255 * msg.timer).textStyle(TEXT_BOLD).fontSize(20).stroke(62).strokeWidth(5)
        style.sortOrder(-5000)
        matrix.push().scale(0.04)
        text(amount.."", 0, 0)
        matrix.pop()
        style.pop()
    end
    
    function msg:update(dt)
        self.y = self.y + dt * 1
        self.timer = self.timer - dt * 1
    end
end

function Gameplay:spawnBoss(args)
    self:spawn(args.enemy, self:randomSpawnPoint()) 
end

function Gameplay:spawnSwarm(args)
    local x, y = self:randomSpawnPoint()
    local dir = vec2()
    dir.x = self.player.entity.x - x
    dir.y = self.player.entity.y - y
    dir:normalize()
    
    for i = 1,args.count do 
        local mon = self:spawn(args.enemy, x + math.random()*0.2-0.1, y + math.random()*0.2-0.1)        
        mon.swarmDir = dir
    end
end

--- Gameplay:beginWave Begin a new wave of enemies
--- wave: table
function Gameplay:beginWave(wave)
    self.waveTimer = 0
    self.spawnTimer = 0
    for i = 1, wave.initial do
        self:spawnEnemy(wave)
    end
end

--- Gameplay:updateWave Update wave logic
--- wave: table
--- dt: number
function Gameplay:updateWave(wave, dt)    
    -- Handle events that occur on certain sections of a given wave
    if wave.events then
        if math.floor(self.waveTimer) == math.ceil(self.waveTimer-dt) then
            local event = wave.events[math.floor(self.waveTimer)]
            if event and math.random() < event.chance then
                self[event[1]](self, event.args)
            end
        end
    end
    
    -- Handle spawning logic (triggered on specific intervals)
    self.spawnTimer = math.max(self.spawnTimer - dt, 0)    
    if self.spawnTimer == 0 then
        self.spawnTimer = wave.spawnRate
        self:spawnEnemy(wave)
    end
end

--- Gameplay:spawnEnemy Spawn a random enemy for a given wave
--- wave: table
function Gameplay:spawnEnemy(wave)
    if self.enemyCount < self.maxEnemies and self.player then
        local enemyType = wave.enemies[math.random(1, #wave.enemies)]
        self:spawn(enemyType, self:randomSpawnPoint())
        self.enemyCount = self.enemyCount + 1
    end
end

--- Gameplay:onEnemyKilled Event called when an enemy dies
--- enemy: Monster
function Gameplay:onEnemyKilled(enemy)
    self.enemyCount = self.enemyCount - 1
end

--- Gameplay:randomSpawnPoint Generate a random spawn point
function Gameplay:randomSpawnPoint()
    if self.player then
        local sx, sy = self:getOrthoSize(WIDTH, HEIGHT)
        local r = math.sqrt(sx*sx + sy*sy) / 2 + 1 + math.random() * 2
        
        local pos = vec2(math.random() * 2.0 - 1.0, math.random() * 2.0 - 1.0)
        pos:normalize()
        pos.x = math.min(math.max(pos.x * r, -sx/2 - 2), sx/2 + 2)
        pos.y = math.min(math.max(pos.y * r, -sy/2 - 2), sy/2 + 2)
        
        pos.x = pos.x + self.player.entity.x
        pos.y = pos.y + self.player.entity.y                
        return pos.x, pos.y
    end
    return nil
end

--- Gameplay:spawnFromTile Spawn an object at a specific tile location
--- objectType: table
--- x: integer
--- y: integer
function Gameplay:spawnFromTile(objectType, x, y)
    local obj = self:spawn(objectType, self.layers[3]:tileToWorld(x, y))

    -- Keep track of the player when they are spawned    
    if objectType.class == Player then
        self.player = obj
        self.playerEntity = self.player.entity
    end
end

--- Gameplay:spawn Spawn an object in world space
--- objectType: table
--- wx: number
--- wy: number
function Gameplay:spawn(objectType, wx, wy)    
    local sx, sy = self:getOrthoSize(WIDTH, HEIGHT)
    if wx < sx/2 or wx > MAP_SIZE+sx/2 or wy < sy/2 or wy > MAP_SIZE+sy/2 then return end
    
    local new = self.scene:entity(objectType.class.name)
    new.x, new.y = wx, wy
    local obj = new:add(objectType.class, self, table.unpack(objectType.args or {}))
    return obj
end

--- Gameplay:onLevelUp Called during levelup event
function Gameplay:onLevelUp()
    -- Push a level up menu
    -- game:push(LevelUpState(self))
end
