--- Player: class
-- A component that handles all player logic
Player = class("Player")

function Player:created(gameplay)
    self.gameplay = gameplay
    self.sprite = tileset.player:sprite()
    self.sprite:anchor(0.5, 0.95)
    self.mob = self.entity:add(Mob, gameplay, self.sprite, PLAYER_CAT)
    self.mob.disableOnKill = true
    self.body = self.mob.body
    self.size = 0.5
    self.dragRadius = 75
    self.stickRadius = 25
    self.aim = vec2(1, 0)
    self.cameraOffset = vec2(0, 0)
    
    self.mob.maxHealth = 100
    self.mob.health = self.mob.maxHealth
    
    self.level = 0
    self.xp = 0
    self.nextLevelXP = self:calculateXP()
    --[[
    self.items = {}
    self.itemMap = {}
    self.weapons = {}
    self.passives = {}
    ]]
    -- Player stats (can be augmented by passive items)
    self.stats = {}
    self.stats.strength = Stat("Strength", 0)
    self.stats.regen = Stat("Regen", 1)
    self.stats.armor = Stat("Armor", 0)
    self.stats.vacuum = Stat("Vacuum", 1)
    self.stats.moveSpeed = Stat("Move Speed", 2)
    self.stats.haste = Stat("Haste", 0)
    self.stats.amount = Stat("Amount", 0)
    
    self.world = self.gameplay.scene.world2d
    
    self.cards = 
    {
        SwordCard(),
        SplitCard(3, 45),
        LaunchCard(),
        TimerCard(),            
        HomingCard()
    }
    
    self.items =
    {
        CardAttack(self, {})
    }
    
end

--- Player:calculateXP Return the XP required for the next level up
function Player:calculateXP()
    return 10 + 20 * self.level
end

--- Player:grantXP Give the player a certain amount of XP
--- xp: number
function Player:grantXP(xp)
    self.xp = self.xp + xp
    while self.xp >= self.nextLevelXP do
        self.xp = self.xp - self.nextLevelXP
        self.level = self.level + 1
        self.nextLevelXP = self:calculateXP()
        self.gameplay:onLevelUp()
        --self:cancelTouches()
    end
end

--- Player:cancelTouches Call during level up to reset touch-based controls
function Player:cancelTouches()
    self.touch = nil
    self.touchStart = nil
end

function Player:onKilledMonster(monster)
end

function math.sign(number)
    return (number > 0 and 1) or (number == 0 and 0) or -1
end

--- Player:touched Handle touch-based controls
--- touch: touch
function Player:touched(touch)
    if self.touch == nil and touch.began then
        self.touch = touch
        self.touchStart = touch.pos
        self.touchPos = touch.pos
        return true
    end
    
    if self.touch and touch.id == self.touch.id then
        if touch.moving then
            self.touchPos = touch.pos
            if self.touchPos:distance(self.touchStart) > (self.dragRadius - self.stickRadius) then
                self.touchPos = (self.touchPos - self.touchStart).normalized * (self.dragRadius - self.stickRadius) + self.touchStart
            end
        elseif touch.ended or touch.cancelled then
            self.touch = nil
            self.touchStart = nil
        end                
    end
end
    
function Player:update(dt)    
    local speed = self.stats.moveSpeed:value()
    
    if self.touch then
        local r = self.dragRadius - self.stickRadius
        self.mob.dir.x = -(self.touchStart.x - self.touchPos.x) / r * speed
        self.mob.dir.y = -(self.touchStart.y - self.touchPos.y) / r * speed
    else 
        self.mob.dir.x = 0
        self.mob.dir.y = 0
    end
    
    local left = key.pressing(key.a) or key.pressing(key.left)
    local right = key.pressing(key.d) or key.pressing(key.right)
    local up = key.pressing(key.w) or key.pressing(key.up)
    local down = key.pressing(key.s) or key.pressing(key.down)
    local inputPressed = left or right or up or down
    
    if inputPressed and self.input == nil then
        self.mob.dir.x, self.mob.dir.y = 0, 0
        if left then self.mob.dir.x = self.mob.dir.x - 1 end
        if right then self.mob.dir.x = self.mob.dir.x + 1 end
        if up then self.mob.dir.y = self.mob.dir.y + 1 end
        if down then self.mob.dir.y = self.mob.dir.y - 1 end        
        self.mob.dir.x = self.mob.dir.x * speed
        self.mob.dir.y = self.mob.dir.y * speed
    end
    
    if self.mob.dir.x ~= 0 or self.mob.dir.y ~= 0 then
        self.aim.x = self.mob.dir.x
        self.aim.y = self.mob.dir.y
    end
    
    -- Gem vacuum
    local radius = self.stats.vacuum:value()
    local items = self.world:queryBoxAll(self.entity.x - radius, self.entity.y - radius, radius * 2, radius * 2, ITEM_CAT)
    for k,v in pairs(items) do
        if v.body.entity then
            local gem = v.body.entity:get(Gem)
            if gem then
                local dir = (self.entity.position - gem.entity.position).normalized * time.delta * 2
                gem.entity.position = gem.entity.position + dir
                gem:collect()
            end
        end
    end
       
    -- Health regen
    self.mob:heal(self.stats.regen:value() * dt)
    
    -- Update items 
    for k,v in pairs(self.items) do
        v:update(dt)
    end
end

function Player:draw()
    -- Draw any active items
    for k,v in pairs(self.items) do
        v:draw()
    end
    
    -- Draw health bar
    style.push().noStroke().fill(9,9,9,128)
    rect(-0.5, -0.25, 1, 0.1)
    local p = self.mob.health / self.mob.maxHealth
    style.fill(255,0,0,128)
    rect(-0.5, -0.25, 1 * p, 0.1)
    style.pop()
end

--- Player:drawUI Draw touch-based UI elements
function Player:drawUI()
    matrix.push().reset()
    if self.touch then
        style.push().noFill().stroke(255).strokeWidth(5).sortOrder(-2000)
        ellipse(self.touchStart, self.dragRadius*2)
        style.fill(255).noStroke()
        ellipse(self.touchPos, self.stickRadius*2)
        style.pop()
    end
    matrix.pop()
end
